#include <iomanip>
#include <iostream>
#include <ostream>
#include <random>
#include <vector>

#include "timer.h"

#define MORTON 0

inline int morton_encode(int x, int y)
{
    auto expand_bits = [](int n) -> int
    {
        // TODO:
        // Use bitwise operations and simple integer
        // constants to expand n's bits so that
        // each bit is interleaved with 0-bits
        // NOTE: we lose 16-bit precision here!
    };

    return expand_bits(x) | (expand_bits(y) << 1);
}

template <typename T>
struct Grid
{
    Grid(int w, int h)
        : width(w)
        , height(h)
        , data(w * h)
    {
    }

    T& operator()(int x, int y)
    {
#if MORTON
        return data[morton_encode(x, y)];
#else
        return data[y * width + x];
#endif
    }

    T const& operator()(int x, int y) const
    {
#if MORTON
        return data[morton_encode(x, y)];
#else
        return data[y * width + x];
#endif
    }

    int width;
    int height;
    std::vector<T> data;
};

int main()
{
    std::cout << std::fixed << std::setprecision(3);

    visionaray::timer t;

    // You may play around with the grid size here
    constexpr int W = 8192;
    constexpr int H = 8192;
    Grid<float> grid(W, H);


    // Random initialize the grid
    std::default_random_engine rng;
    std::uniform_real_distribution<float> dist;

    for (int y = 0; y < H; ++y)
    {
        for (int x = 0; x < W; ++x)
        {
            grid(x, y) = dist(rng);
        }
    }
    std::cout << "Initialization: " << t.elapsed() << '\n';
    t.reset();

    Grid<float> filtered(W, H);

    float filter[3][3] = {
       {  1.f,  2.f,  1.f },
       {  0.f,  0.f,  0.f },
       { -1.f, -2.f, -1.f }
       };

    // Apply filter from top to bottom
    for (int y = 1; y < H - 1; ++y)
    {
        for (int x = 1; x < W - 1; ++x)
        {
            filtered(x, y) = grid(x - 1, y - 1) * filter[0][0] 
                           + grid(x,     y - 1) * filter[1][0]
                           + grid(x + 1, y - 1) * filter[2][0]
                           + grid(x - 1, y    ) * filter[0][1]
                           + grid(x,     y    ) * filter[1][1]
                           + grid(x + 1, y    ) * filter[2][1]
                           + grid(x - 1, y + 1) * filter[0][2]
                           + grid(x,     y + 1) * filter[1][2]
                           + grid(x + 1, y + 1) * filter[2][2];
        }
    }
    std::cout << "Horizontal filter: " << t.elapsed() << '\n';

    t.reset();

    // Apply filter from left to right
    for (int x = 1; x < W - 1; ++x)
    {
        for (int y = 1; y < H - 1; ++y)
        {
            filtered(x, y) = grid(x - 1, y - 1) * filter[0][0] 
                           + grid(x,     y - 1) * filter[1][0]
                           + grid(x + 1, y - 1) * filter[2][0]
                           + grid(x - 1, y    ) * filter[0][1]
                           + grid(x,     y    ) * filter[1][1]
                           + grid(x + 1, y    ) * filter[2][1]
                           + grid(x - 1, y + 1) * filter[0][2]
                           + grid(x,     y + 1) * filter[1][2]
                           + grid(x + 1, y + 1) * filter[2][2];
        }
    }

    std::cout << "Vertical filter: " << t.elapsed() << '\n';

    // Make sure we use the filtered data so the compiler
    // cannot optimize the whole program away
    return filtered(dist(rng) * W - 1, dist(rng) * H - 1);
}
